{-# LANGUAGE ForeignFunctionInterface #-}
-- link with -lhdf5 -lhdf5_hl
module HDF5 (module HDF5.Base,
             emptyListOf,
             append,
             readInto)
where

import Foreign
import Foreign.C
import Data.Array.Storable
import Data.Complex (Complex((:+)))
import Control.Exception (try,tryJust,bracket,SomeException)
import Control.Monad (guard,liftM,when)
import System.IO.Error (isAlreadyExistsError,isDoesNotExistError)

import HDF5.Base
import qualified HDF5.File as H5F
import qualified HDF5.Lite as H5LT
import qualified HDF5.Type as H5T
import qualified HDF5.Property as H5P
import qualified HDF5.Dataset as H5D
import qualified HDF5.Dataspace as H5S
import qualified HDF5.Error as H5E
import qualified HDF5.Library as H5

emptyListOf :: Lid -> String -> Tid -> Hsize -> IO Did
emptyListOf loc name typeId (Hsize elemSize) =
  bracket (H5S.create_simple (Hsize $ elemSize ++ [0]) (Hsize $ elemSize ++ [maxBound]))
          H5S.close $ \emptySpace ->
    bracket (H5P.create H5P.dataset_create)
             H5P.close $ \chunkPropList -> do
      H5P.set_chunk chunkPropList (Hsize $ elemSize ++ [1])
      H5D.create loc name typeId emptySpace H5P._default chunkPropList H5P._default

append :: (Storable a,H5T.Typed a) => Did -> StorableArray Hsize a -> IO ()
append dset elem =
  H5S.withSpaceOf dset $ \origSpace -> do
    (rank,Hsize dims,_) <- H5S.get_simple_extent_dims origSpace
    elemDims <- dimensions `liftM` getBounds elem
    when (length dims /= length elemDims + 1) $
      ioError $ userError $ "not a list of "++show elemDims
    let currLength = last dims
    H5D.set_extent dset (Hsize $ elemDims ++ [currLength + 1])
    H5S.withSpaceOf dset $ \biggerSpace -> do
      H5S.select_hyperslab biggerSpace H5S.select_set
        (Hsize $ replicate (rank-1) 0 ++ [currLength]) Nothing (Hsize $ elemDims ++ [1]) Nothing
      H5S.withSimple (Hsize elemDims) $ \memSpace ->
        withStorableArray elem $ \elemP ->
        H5D.write dset memSpace biggerSpace elemP

readInto :: (Storable a,H5T.Typed a) => Lid -> String -> StorableArray Hsize a -> IO ()
readInto loc name arr = do
  bounds <- getBounds arr
  let dims = dimensions bounds
      rank = length dims
  H5S.withSimple (Hsize dims) $ \memSpace ->
    H5D.withDataset loc name $ \dset ->
    H5S.withSpaceOf dset $ \dspace -> do
      H5S.select_hyperslab dspace H5S.select_set
        (Hsize $ replicate rank 0) Nothing (Hsize dims) Nothing
      withStorableArray arr $ \arrP ->
        H5D.read dset memSpace dspace arrP

writeTest :: Lid -> IO String
writeTest t = do
  scalar <- newListArray (Hsize [],Hsize []) [(1::CFloat)..]
  vect <- newListArray (Hsize [1],Hsize [5]) [(1::CFloat)..]
  row <- newListArray (Hsize [1,1],Hsize [1,5]) [(1::CFloat)..]
  col <- newListArray (Hsize [1,1],Hsize [4,1]) [(1::CFloat)..]
  mat <- newListArray (Hsize [1,1],Hsize [3,4]) [(1::CFloat)..]
  multi <- newListArray (Hsize [1,1,1],Hsize [2,3,4]) [(1::CFloat)..]
  as <- getAssocs multi
  bounds <- getBounds multi
  print $ dimensions bounds
  print as
  H5LT.make_dataset t "scalar" scalar
  H5LT.make_dataset t "vect" vect
  H5LT.make_dataset t "row" row
  H5LT.make_dataset t "col" col
  H5LT.make_dataset t "mat" mat
  H5LT.make_dataset t "multi" multi
  H5LT.make_scalar t "c" (1:+(2::CFloat))
  bracket (emptyListOf t "empty" H5T.native_int (Hsize [2,3]))
          H5D.close $ \dset -> do
    plane1 <- newListArray (Hsize [1,1],Hsize [2,3]) [(1::CInt)..]
    append dset plane1
    plane2 <- newListArray (Hsize [1,1],Hsize [2,3]) [(2::CInt)..]
    append dset plane2
    return "written"

withNewFile :: String -> (Lid -> IO a) -> IO a
withNewFile filename =
  bracket (H5E.withMuteErrorStack (H5F.create filename H5F.EXCL))
          H5F.close

readTest :: Lid -> IO CFloat
readTest t = do
  matPart <- newArray_ (Hsize [1,1],Hsize [2,3])
  res <- try $ readInto t "matt" matPart
  case res of
    Left e   -> print (e :: SomeException)
    Right () -> putStrLn "WTF?"
  readInto t "mat" matPart
  sum `liftM` getElems matPart

withExistingFile :: String -> (Lid -> IO a) -> IO a
withExistingFile filename =
  bracket (H5E.withMuteErrorStack (H5F.open filename H5F.RDONLY))
          H5F.close

test = do
  res <- tryJust (guard . isDoesNotExistError) test_reading
  case res of
    Left () -> putStrLn "Does not exist yet."
    Right r -> putStrLn $ "Read result: " ++ show r
  res <- tryJust (guard . isAlreadyExistsError) (withNewFile testFile writeTest)
  case res of
    Left () -> putStrLn "Already exists." >> test_reading >>= print
    Right r -> putStrLn $ "Write result: " ++ r
  where testFile = "test.h5"
        test_reading = withExistingFile testFile readTest
