{-# LANGUAGE ForeignFunctionInterface #-}
module HDF5.Dataset (create,
                     open,
                     close,
                     withDataset,
                     get_space,
                     set_extent,
                     read,
                     write)
where

import Prelude hiding (read)
import Foreign
import Foreign.C
import Control.Exception (bracket)

import HDF5.Base
import qualified HDF5.Error as H5E
import qualified HDF5.Property as H5P
import qualified HDF5.Type as H5T

foreign import ccall unsafe "hdf5.h H5Dcreate2"
--            loc_id   name    dtype  space   lcpl   dcpl   dapl
  create2_ :: Lid -> CString -> Tid -> Sid -> Pid -> Pid -> Pid -> IO Did

create :: Lid -> String -> Tid -> Sid -> Pid -> Pid -> Pid -> IO Did
create loc name dtype space lcpl dcpl dapl =
  withCString name $ \cname ->
    H5E.throwIfErr ("creating dataset "++name) $
      create2_ loc cname dtype space lcpl dcpl dapl

foreign import ccall unsafe "hdf5.h H5Dclose"
  close_ :: Did -> IO Herr

close :: Did -> IO ()
close =
  H5E.throwIfErr_ "closing dataset" . close_

foreign import ccall unsafe "hdf5.h H5Dget_space"
  get_space_ :: Did -> IO Sid

get_space :: Did -> IO Sid
get_space =
  H5E.throwIfErr "getting dataspace" . get_space_

foreign import ccall unsafe "hdf5.h H5Dset_extent"
  set_extent_ :: Did -> Ptr CULLong -> IO Herr

set_extent :: Did -> Hsize -> IO ()
set_extent dset (Hsize hs) =
  withArray (reverse hs) $ \hsP ->
  H5E.throwIfErr_ ("setting extent to "++show hs) $ set_extent_ dset hsP

foreign import ccall unsafe "hdf5.h H5Dopen2"
--         loc_id    name   dapl_id
  open2_ :: Lid -> CString -> Pid -> IO Did

open :: Lid -> String -> IO Did
open loc name =
  withCString name $ \cname ->
  H5E.throwIfErr ("opening dataset "++name) $ open2_ loc cname H5P._default

withDataset :: Lid -> String -> (Did -> IO a) -> IO a
withDataset loc name = bracket (open loc name) close

foreign import ccall unsafe "hdf5.h H5Dread"
--        dset mem_type mem_space file_space xfer_pl buf
  read_ :: Did -> Tid -> Sid ->   Sid ->     Pid -> Ptr a -> IO Herr

read :: H5T.Typed a => Did -> Sid -> Sid -> Ptr a -> IO ()
read dset memSpace fileSpace buf =
  H5T.withTypeP buf $ \memType ->
  H5E.throwIfErr_ "reading dataset" $
    read_ dset memType memSpace fileSpace H5P._default buf

foreign import ccall unsafe "hdf5.h H5Dwrite"
--         dset mem_type mem_space file_space xfer_pl buf
  write_ :: Did -> Tid -> Sid ->   Sid ->     Pid -> Ptr a -> IO Herr

write :: H5T.Typed a => Did -> Sid -> Sid -> Ptr a -> IO ()
write dset memSpace fileSpace buf =
  H5T.withTypeP buf $ \memType ->
  H5E.throwIfErr_ "writing dataset" $
    write_ dset memType memSpace fileSpace H5P._default buf