{-# LANGUAGE ForeignFunctionInterface #-}
module HDF5.Dataspace (create_simple,
                       close,
                       get_simple_extent_ndims,
                       get_simple_extent_dims,
                       select_hyperslab,
                       select_set,
                       withSpaceOf,
                       withSimple)
where

import Foreign
import Foreign.C
import Control.Exception (bracket)
import Control.Monad (unless)
import Data.Maybe (catMaybes)

import HDF5.Base
import qualified HDF5.Dataset as H5D
import qualified HDF5.Error as H5E

foreign import ccall unsafe "hdf5.h H5Screate_simple"
--                  rank       dims          maxdims
  create_simple_ :: CInt -> Ptr CULLong -> Ptr CULLong -> IO Sid

create_simple :: Hsize -> Hsize -> IO Sid
create_simple (Hsize dims) (Hsize maxdims) =
  withArrayLen (reverse dims) $ \dimsLen dimsP ->
  withArrayLen (reverse maxdims) $ \maxdimsLen maxdimsP ->
  if (dimsLen /= maxdimsLen)
  then ioError $ userError $ "Incompatible actual and maximal ranks: "++show dimsLen++"/="++show maxdimsLen
  else H5E.throwIfErr ("creating simple dataspace of rank "++show dimsLen++", dims "++show dims++", maxdims "++show maxdims) $
         create_simple_ (fromIntegral dimsLen) dimsP maxdimsP

foreign import ccall unsafe "hdf5.h H5Sclose"
  close_ :: Sid -> IO Herr

close :: Sid -> IO ()
close =
  H5E.throwIfErr_ "closing dataset" . close_

foreign import ccall unsafe "hdf5.h H5Sget_simple_extent_ndims"
  get_simple_extent_ndims_ :: Sid -> IO CInt

get_simple_extent_ndims :: Sid -> IO Int
get_simple_extent_ndims dspace = do
  rank <- get_simple_extent_ndims_ dspace
  if rank < 0
    then ioError $ userError "getting simple extent ndims"
    else return $ fromIntegral rank

foreign import ccall unsafe "hdf5.h H5Sget_simple_extent_dims"
  get_simple_extent_dims_ :: Sid -> Ptr CULLong -> Ptr CULLong -> IO Herr

get_simple_extent_dims :: Sid -> IO (Int,Hsize,Hsize)
get_simple_extent_dims dspace = do
  rank <- get_simple_extent_ndims dspace
  allocaArray rank $ \dimsP ->
    allocaArray rank $ \maxdimsP -> do
      H5E.throwIfErr_ "getting simple extent dims" $ get_simple_extent_dims_ dspace dimsP maxdimsP
      dims <- peekArray rank dimsP
      maxdims <- peekArray rank maxdimsP
      return (rank,Hsize $ reverse dims,Hsize $ reverse maxdims)

newtype Operation = Operation CInt

select_set     = Operation 0      -- Select "set" operation
select_or      = Operation 1      -- Binary "or" operation for hyperslabs
select_and     = Operation 2      -- Binary "and" operation for hyperslabs
select_xor     = Operation 3      -- Binary "xor" operation for hyperslabs
select_notb    = Operation 4      -- Binary "not" operation for hyperslabs
select_nota    = Operation 5      -- Binary "not" operation for hyperslabs
select_append  = Operation 6      -- Append elements to end of point selection
select_prepend = Operation 7      -- Prepend elements to beginning of point selection

foreign import ccall unsafe "hdf5.h H5Sselect_hyperslab"
--                    space      op           start         stride
  select_hyperslab_ :: Sid -> Operation -> Ptr CULLong -> Ptr CULLong ->
--                        count           block
                       Ptr CULLong -> Ptr CULLong -> IO Herr

select_hyperslab :: Sid -> Operation -> Hsize -> Maybe Hsize -> Hsize -> Maybe Hsize -> IO ()
select_hyperslab space op (Hsize start) stride' (Hsize count) block' = do
  rank <- get_simple_extent_ndims space
  unless (all (== fromIntegral rank) lengths) $
    ioError $ userError $ "Inconsistent ranks in hyperslab selection: "++show rank++
                          "against "++unwords (map show lengths)
  withArray (reverse start) $ \startP ->
    maybeWith withArray (fmap reverse stride) $ \strideP ->
    withArray (reverse count) $ \countP ->
    maybeWith withArray (fmap reverse block) $ \blockP ->
    H5E.throwIfErr_ "selecting hyperslab" $
      select_hyperslab_ space op startP strideP countP blockP
  where lengths              = map length $ [start,count] ++ catMaybes opLen
        opLen@[stride,block] = map (fmap (\(Hsize x) -> x)) [stride',block']

withSpaceOf :: Did -> (Sid -> IO a) -> IO a
withSpaceOf dset = bracket (H5D.get_space dset) close

withSimple :: Hsize -> (Sid -> IO a) -> IO a
withSimple hs = bracket (create_simple hs hs) close
