{-# LANGUAGE ForeignFunctionInterface #-}
module HDF5.Error (withMuteErrorStack,
                   throwIfErr,
                   throwIfErr_)
where

import Foreign
import Foreign.C
import Control.Exception (bracket,bracket_,bracketOnError)
import Control.Monad (liftM)
import System.IO (hPutStrLn,stderr)

import HDF5.Base
import HDF5.Platform

_default = Eid 0

-- typedef herr_t (*H5E_auto2_t)(hid_t estack, void *client_data);
type Auto a = FunPtr (Eid -> Ptr a -> IO Herr)

foreign import ccall unsafe "hdf5.h H5Eget_auto2"
  get_auto_ :: Eid -> Ptr (Auto a) -> Ptr (Ptr a) -> IO Herr

get_auto :: Eid -> Ptr (Auto a) -> Ptr (Ptr a) -> IO ()
get_auto estack autoPP client_dataPP =
  throwIfErr_ "getting auto" $ get_auto_ estack autoPP client_dataPP

foreign import ccall unsafe "hdf5.h H5Eset_auto2"
  set_auto_ :: Eid -> Auto a -> Ptr a -> IO Herr

set_auto :: Eid -> Auto a -> Ptr a -> IO ()
set_auto estack autoP client_dataP =
  throwIfErr_ "setting auto" $ set_auto_ estack autoP client_dataP

withMuteErrorStack :: IO a -> IO a
withMuteErrorStack action =
  alloca $ \funcPP ->
  alloca $ \client_dataPP ->
  bracket_ (do get_auto _default funcPP client_dataPP
               set_auto _default nullFunPtr nullPtr)
           (do funcP <- peek funcPP
               client_dataP <- peek client_dataPP
               set_auto _default funcP client_dataP)
           action

type WalkFunc a = CUInt -> Ptr Error -> Ptr a -> IO Herr

foreign import ccall "wrapper" mkWalkFunc :: WalkFunc a -> IO (FunPtr (WalkFunc a))

withWalkFunc :: WalkFunc a -> (FunPtr (WalkFunc a) -> IO b) -> IO b
withWalkFunc walkFunc =
  bracket (mkWalkFunc walkFunc) freeHaskellFunPtr

newtype Direction = Direction CInt

walk_upward   = Direction 0 -- begin deep, end at API function
walk_downward = Direction 1 -- begin at API function, end deep

foreign import ccall safe "hdf5.h H5Ewalk2"
  walk2_ :: Eid -> Direction -> FunPtr (WalkFunc a) -> Ptr a -> IO Herr

printErrorStackElement :: WalkFunc ()
printErrorStackElement n errP _ = do
  putStr $ show n
  putStr ": "
  peek errP >>= print
  return (Herr 0)

printErrorStack :: IO ()
printErrorStack =
  withWalkFunc printErrorStackElement $ \walkFunc ->
    throwIfErr_ "walking stack" $ walk2_ _default walk_downward walkFunc nullPtr

getFirstErrorStackElement :: WalkFunc (Ptr Error)
getFirstErrorStackElement 0 errP errPP = do
  poke errPP errP
  return $ Herr 0 -- no way to stop walking without provoking an error
getFirstErrorStackElement _ _    _ =
  return $ Herr 0 -- so just walk along

-- getPrincipalError == head `liftM` getErrorStack, but much-much simpler
getPrincipalError :: IO Error
getPrincipalError =
  withWalkFunc getFirstErrorStackElement $ \walkFunc ->
    with nullPtr $ \errPP -> do
      throwIfErr_ "getting principal error" $ walk2_ _default walk_upward walkFunc errPP
      throwIfNull "no error on stack" (peek errPP) >>= peek

deRefAndFreeStablePtrP :: Ptr (Ptr ()) -> IO a
deRefAndFreeStablePtrP aPP = do
  aSP <- castPtrToStablePtr `liftM` peek aPP
  a <- deRefStablePtr aSP
  freeStablePtr aSP
  return a

getErrorStackElement :: WalkFunc (Ptr ())
getErrorStackElement _ errP errPListPP = do
  errPList <- deRefAndFreeStablePtrP errPListPP
  errPListSP <- newStablePtr (errP:errPList)
  poke errPListPP (castStablePtrToPtr errPListSP)
  return $ Herr 0

getErrorStack :: IO [Error]
getErrorStack =
  withWalkFunc getErrorStackElement $ \walkFunc ->
    bracketOnError (newStablePtr [])
                   freeStablePtr $ \errPListSP ->
      with (castStablePtrToPtr errPListSP) $ \errPListPP -> do
        throwIfErr_ "getting error stack" $ walk2_ _default walk_downward walkFunc errPListPP
        errPList <- deRefAndFreeStablePtrP errPListPP
        mapM peek errPList

throwError :: String -> IO a
throwError loc = do
  errno <- getErrno -- save it ASAP
  errStack <- getErrorStack
  ioError $ case errStack of
    []      -> userError $ loc ++ ": (empty error stack)"
    (err:_) -> if maj_num err == major_internal && min_num err == minor_syserrstr
               then errnoToIOError loc errno Nothing (Just $ func_name err)
               else userError $ loc ++ ": " ++ show errStack

throwIfErr :: Hid a => String -> IO a -> IO a
throwIfErr loc action = do
  res <- action
  if isFailure res then throwError loc else return res

throwIfErr_ :: String -> IO Herr -> IO ()
throwIfErr_ loc action = do
  Herr x <- action
  if (x < 0) then throwError loc else return ()

peekEid = Eid . unsafePerformIO . peek

-- Major error codes

major_func       = peekEid major_func_g       -- Function entry/exit
major_storage    = peekEid major_storage_g    -- Data storage
major_file       = peekEid major_file_g       -- File accessability
major_sohm       = peekEid major_sohm_g       -- Shared Object Header Messages
major_sym        = peekEid major_sym_g        -- Symbol table
major_vfl        = peekEid major_vfl_g        -- Virtual File Layer
major_internal   = peekEid major_internal_g   -- Internal error (too specific to document in detail)
major_btree      = peekEid major_btree_g      -- B-Tree node
major_reference  = peekEid major_reference_g  -- References
major_dataspace  = peekEid major_dataspace_g  -- Dataspace
major_resource   = peekEid major_resource_g   -- Resource unavailable
major_plist      = peekEid major_plist_g      -- Property lists
major_link       = peekEid major_link_g       -- Links
major_datatype   = peekEid major_datatype_g   -- Datatype
major_rs         = peekEid major_rs_g         -- Reference Counted Strings
major_heap       = peekEid major_heap_g       -- Heap
major_ohdr       = peekEid major_ohdr_g       -- Object header
major_atom       = peekEid major_atom_g       -- Object atom
major_attr       = peekEid major_attr_g       -- Attribute
major_none_major = peekEid major_none_major_g -- No error
major_io         = peekEid major_io_g         -- Low-level I/O
major_slist      = peekEid major_slist_g      -- Skip Lists
major_efl        = peekEid major_efl_g        -- External file list
major_tst        = peekEid major_tst_g        -- Ternary Search Trees
major_args       = peekEid major_args_g       -- Invalid arguments to routine
major_error      = peekEid major_error_g      -- Error API
major_pline      = peekEid major_pline_g      -- Data filters
major_fspace     = peekEid major_fspace_g     -- Free Space Manager
major_cache      = peekEid major_cache_g      -- Object cache

-- Minor error codes

-- Generic low-level file I/O errors
minor_seekerror  = peekEid minor_seekerror_g  -- Seek failed
minor_readerror  = peekEid minor_readerror_g  -- Read failed
minor_writeerror = peekEid minor_writeerror_g -- Write failed
minor_closeerror = peekEid minor_closeerror_g -- Close failed
minor_overflow   = peekEid minor_overflow_g   -- Address overflowed
minor_fcntl      = peekEid minor_fcntl_g      -- File control (fcntl) failed

-- Resource errors
minor_nospace       = peekEid minor_nospace_g       -- No space available for allocation
minor_cantalloc     = peekEid minor_cantalloc_g     -- Can't allocate space
minor_cantcopy      = peekEid minor_cantcopy_g      -- Unable to copy object
minor_cantfree      = peekEid minor_cantfree_g      -- Unable to free object
minor_alreadyexists = peekEid minor_alreadyexists_g -- Object already exists
minor_cantlock      = peekEid minor_cantlock_g      -- Unable to lock object
minor_cantunlock    = peekEid minor_cantunlock_g    -- Unable to unlock object
minor_cantgc        = peekEid minor_cantgc_g        -- Unable to garbage collect
minor_cantgetsize   = peekEid minor_cantgetsize_g   -- Unable to compute size
minor_objopen       = peekEid minor_objopen_g       -- Object is already open

-- Heap errors
minor_cantrestore = peekEid minor_cantrestore_g -- Can't restore condition
minor_cantcompute = peekEid minor_cantcompute_g -- Can't compute value
minor_cantextend  = peekEid minor_cantextend_g  -- Can't extend heap's space
minor_cantattach  = peekEid minor_cantattach_g  -- Can't attach object
minor_cantupdate  = peekEid minor_cantupdate_g  -- Can't update object
minor_cantoperate = peekEid minor_cantoperate_g -- Can't operate on object

-- Function entry/exit interface errors
minor_cantinit    = peekEid minor_cantinit_g    -- Unable to initialize object
minor_alreadyinit = peekEid minor_alreadyinit_g -- Object already initialized
minor_cantrelease = peekEid minor_cantrelease_g -- Unable to release object

-- Property list errors
minor_cantget  = peekEid minor_cantget_g  -- Can't get value
minor_cantset  = peekEid minor_cantset_g  -- Can't set value
minor_dupclass = peekEid minor_dupclass_g -- Duplicate class name in parent class

-- Free space errors
minor_cantmerge  = peekEid minor_cantmerge_g  -- Can't merge objects
minor_cantrevive = peekEid minor_cantrevive_g -- Can't revive object
minor_cantshrink = peekEid minor_cantshrink_g -- Can't shrink container

-- Object header related errors
minor_linkcount  = peekEid minor_linkcount_g  -- Bad object header link count
minor_version    = peekEid minor_version_g    -- Wrong version number
minor_alignment  = peekEid minor_alignment_g  -- Alignment error
minor_badmesg    = peekEid minor_badmesg_g    -- Unrecognized message
minor_cantdelete = peekEid minor_cantdelete_g -- Can't delete message
minor_baditer    = peekEid minor_baditer_g    -- Iteration failed
minor_cantpack   = peekEid minor_cantpack_g   -- Can't pack messages
minor_cantreset  = peekEid minor_cantreset_g  -- Can't reset object

-- System level errors
minor_syserrstr = peekEid minor_syserrstr_g -- System error message

-- I/O pipeline errors
minor_nofilter   = peekEid minor_nofilter_g   -- Requested filter is not available
minor_callback   = peekEid minor_callback_g   -- Callback failed
minor_canapply   = peekEid minor_canapply_g   -- Error from filter 'can apply' callback
minor_setlocal   = peekEid minor_setlocal_g   -- Error from filter 'set local' callback
minor_noencoder  = peekEid minor_noencoder_g  -- Filter present but encoding disabled
minor_cantfilter = peekEid minor_cantfilter_g -- Filter operation failed

-- Group related errors
minor_cantopenobj  = peekEid minor_cantopenobj_g  -- Can't open object
minor_cantcloseobj = peekEid minor_cantcloseobj_g -- Can't close object
minor_complen      = peekEid minor_complen_g      -- Name component is too long
minor_path         = peekEid minor_path_g         -- Problem with path to object

-- No error
minor_none_minor = peekEid minor_none_minor_g -- No error

-- File accessability errors
minor_fileexists    = peekEid minor_fileexists_g    -- File already exists
minor_fileopen      = peekEid minor_fileopen_g      -- File already open
minor_cantcreate    = peekEid minor_cantcreate_g    -- Unable to create file
minor_cantopenfile  = peekEid minor_cantopenfile_g  -- Unable to open file
minor_cantclosefile = peekEid minor_cantclosefile_g -- Unable to close file
minor_nothdf5       = peekEid minor_nothdf5_g       -- Not an HDF5 file
minor_badfile       = peekEid minor_badfile_g       -- Bad file ID accessed
minor_truncated     = peekEid minor_truncated_g     -- File has been truncated
minor_mount         = peekEid minor_mount_g         -- File mount error

-- Object atom related errors
minor_badatom      = peekEid minor_badatom_g      -- Unable to find atom information (already closed?)
minor_badgroup     = peekEid minor_badgroup_g     -- Unable to find ID group information
minor_cantregister = peekEid minor_cantregister_g -- Unable to register new atom
minor_cantinc      = peekEid minor_cantinc_g      -- Unable to increment reference count
minor_cantdec      = peekEid minor_cantdec_g      -- Unable to decrement reference count
minor_noids        = peekEid minor_noids_g        -- Out of IDs for group

-- Cache related errors
minor_cantflush     = peekEid minor_cantflush_g     -- Unable to flush data from cache
minor_cantserialize = peekEid minor_cantserialize_g -- Unable to serialize data from cache
minor_cantload      = peekEid minor_cantload_g      -- Unable to load metadata into cache
minor_protect       = peekEid minor_protect_g       -- Protected metadata error
minor_notcached     = peekEid minor_notcached_g     -- Metadata not currently cached
minor_system        = peekEid minor_system_g        -- Internal error detected
minor_cantins       = peekEid minor_cantins_g       -- Unable to insert metadata into cache
minor_cantrename    = peekEid minor_cantrename_g    -- Unable to rename metadata
minor_cantprotect   = peekEid minor_cantprotect_g   -- Unable to protect metadata
minor_cantunprotect = peekEid minor_cantunprotect_g -- Unable to unprotect metadata
minor_cantpin       = peekEid minor_cantpin_g       -- Unable to pin cache entry
minor_cantunpin     = peekEid minor_cantunpin_g     -- Unable to un-pin cache entry
minor_cantmarkdirty = peekEid minor_cantmarkdirty_g -- Unable to mark a pinned entry as dirty
minor_cantdirty     = peekEid minor_cantdirty_g     -- Unable to mark metadata as dirty
minor_cantexpunge   = peekEid minor_cantexpunge_g   -- Unable to expunge a metadata cache entry
minor_cantresize    = peekEid minor_cantresize_g    -- Unable to resize a metadata cache entry

-- Link related errors
minor_traverse      = peekEid minor_traverse_g      -- Link traversal failure
minor_nlinks        = peekEid minor_nlinks_g        -- Too many soft links in path
minor_notregistered = peekEid minor_notregistered_g -- Link class not registered
minor_cantmove      = peekEid minor_cantmove_g      -- Move callback returned error
minor_cantsort      = peekEid minor_cantsort_g      -- Can't sort objects

-- Parallel MPI errors
minor_mpi       = peekEid minor_mpi_g       -- Some MPI function failed
minor_mpierrstr = peekEid minor_mpierrstr_g -- MPI Error String
minor_cantrecv  = peekEid minor_cantrecv_g  -- Can't receive data

-- Dataspace errors
minor_cantclip    = peekEid minor_cantclip_g    -- Can't clip hyperslab region
minor_cantcount   = peekEid minor_cantcount_g   -- Can't count elements
minor_cantselect  = peekEid minor_cantselect_g  -- Can't select hyperslab
minor_cantnext    = peekEid minor_cantnext_g    -- Can't move to next iterator location
minor_badselect   = peekEid minor_badselect_g   -- Invalid selection
minor_cantcompare = peekEid minor_cantcompare_g -- Can't compare objects

-- Argument errors
minor_uninitialized = peekEid minor_uninitialized_g -- Information is uinitialized
minor_unsupported   = peekEid minor_unsupported_g   -- Feature is unsupported
minor_badtype       = peekEid minor_badtype_g       -- Inappropriate type
minor_badrange      = peekEid minor_badrange_g      -- Out of range
minor_badvalue      = peekEid minor_badvalue_g      -- Bad value

-- B-tree related errors
minor_notfound         = peekEid minor_notfound_g         -- Object not found
minor_exists           = peekEid minor_exists_g           -- Object already exists
minor_cantencode       = peekEid minor_cantencode_g       -- Unable to encode value
minor_cantdecode       = peekEid minor_cantdecode_g       -- Unable to decode value
minor_cantsplit        = peekEid minor_cantsplit_g        -- Unable to split node
minor_cantredistribute = peekEid minor_cantredistribute_g -- Unable to redistribute records
minor_cantswap         = peekEid minor_cantswap_g         -- Unable to swap records
minor_cantinsert       = peekEid minor_cantinsert_g       -- Unable to insert object
minor_cantlist         = peekEid minor_cantlist_g         -- Unable to list node
minor_cantmodify       = peekEid minor_cantmodify_g       -- Unable to modify record
minor_cantremove       = peekEid minor_cantremove_g       -- Unable to remove object

-- Datatype conversion errors
minor_cantconvert = peekEid minor_cantconvert_g -- Can't convert datatypes
minor_badsize     = peekEid minor_badsize_g     -- Bad size for object

-- Raw address import for the above
foreign import ccall unsafe "& H5E_DATASET_g"    major_dataset_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_FUNC_g"       major_func_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_STORAGE_g"    major_storage_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_FILE_g"       major_file_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_SOHM_g"       major_sohm_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_SYM_g"        major_sym_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_VFL_g"        major_vfl_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_INTERNAL_g"   major_internal_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_BTREE_g"      major_btree_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_REFERENCE_g"  major_reference_g  :: Ptr CInt
foreign import ccall unsafe "& H5E_DATASPACE_g"  major_dataspace_g  :: Ptr CInt
foreign import ccall unsafe "& H5E_RESOURCE_g"   major_resource_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_PLIST_g"      major_plist_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_LINK_g"       major_link_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_DATATYPE_g"   major_datatype_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_RS_g"         major_rs_g         :: Ptr CInt
foreign import ccall unsafe "& H5E_HEAP_g"       major_heap_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_OHDR_g"       major_ohdr_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_ATOM_g"       major_atom_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_ATTR_g"       major_attr_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_NONE_MAJOR_g" major_none_major_g :: Ptr CInt
foreign import ccall unsafe "& H5E_IO_g"         major_io_g         :: Ptr CInt
foreign import ccall unsafe "& H5E_SLIST_g"      major_slist_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_EFL_g"        major_efl_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_TST_g"        major_tst_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_ARGS_g"       major_args_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_ERROR_g"      major_error_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_PLINE_g"      major_pline_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_FSPACE_g"     major_fspace_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CACHE_g"      major_cache_g      :: Ptr CInt

foreign import ccall unsafe "& H5E_SEEKERROR_g"     minor_seekerror_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_READERROR_g"     minor_readerror_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_WRITEERROR_g"    minor_writeerror_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CLOSEERROR_g"    minor_closeerror_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_OVERFLOW_g"      minor_overflow_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_FCNTL_g"         minor_fcntl_g         :: Ptr CInt

foreign import ccall unsafe "& H5E_NOSPACE_g"       minor_nospace_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTALLOC_g"     minor_cantalloc_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCOPY_g"      minor_cantcopy_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTFREE_g"      minor_cantfree_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_ALREADYEXISTS_g" minor_alreadyexists_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTLOCK_g"      minor_cantlock_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTUNLOCK_g"    minor_cantunlock_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTGC_g"        minor_cantgc_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTGETSIZE_g"   minor_cantgetsize_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_OBJOPEN_g"       minor_objopen_g       :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTRESTORE_g"   minor_cantrestore_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCOMPUTE_g"   minor_cantcompute_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTEXTEND_g"    minor_cantextend_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTATTACH_g"    minor_cantattach_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTUPDATE_g"    minor_cantupdate_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTOPERATE_g"   minor_cantoperate_g   :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTINIT_g"      minor_cantinit_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_ALREADYINIT_g"   minor_alreadyinit_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTRELEASE_g"   minor_cantrelease_g   :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTGET_g"       minor_cantget_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSET_g"       minor_cantset_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_DUPCLASS_g"      minor_dupclass_g      :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTMERGE_g"     minor_cantmerge_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTREVIVE_g"    minor_cantrevive_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSHRINK_g"    minor_cantshrink_g    :: Ptr CInt

foreign import ccall unsafe "& H5E_LINKCOUNT_g"     minor_linkcount_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_VERSION_g"       minor_version_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_ALIGNMENT_g"     minor_alignment_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_BADMESG_g"       minor_badmesg_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTDELETE_g"    minor_cantdelete_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_BADITER_g"       minor_baditer_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTPACK_g"      minor_cantpack_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTRESET_g"     minor_cantreset_g     :: Ptr CInt

foreign import ccall unsafe "& H5E_SYSERRSTR_g"     minor_syserrstr_g     :: Ptr CInt

foreign import ccall unsafe "& H5E_NOFILTER_g"      minor_nofilter_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CALLBACK_g"      minor_callback_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANAPPLY_g"      minor_canapply_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_SETLOCAL_g"      minor_setlocal_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_NOENCODER_g"     minor_noencoder_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTFILTER_g"    minor_cantfilter_g    :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTOPENOBJ_g"   minor_cantopenobj_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCLOSEOBJ_g"  minor_cantcloseobj_g  :: Ptr CInt
foreign import ccall unsafe "& H5E_COMPLEN_g"       minor_complen_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_PATH_g"          minor_path_g          :: Ptr CInt

foreign import ccall unsafe "& H5E_NONE_MINOR_g"    minor_none_minor_g    :: Ptr CInt

foreign import ccall unsafe "& H5E_FILEEXISTS_g"    minor_fileexists_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_FILEOPEN_g"      minor_fileopen_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCREATE_g"    minor_cantcreate_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTOPENFILE_g"  minor_cantopenfile_g  :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCLOSEFILE_g" minor_cantclosefile_g :: Ptr CInt
foreign import ccall unsafe "& H5E_NOTHDF5_g"       minor_nothdf5_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_BADFILE_g"       minor_badfile_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_TRUNCATED_g"     minor_truncated_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_MOUNT_g"         minor_mount_g         :: Ptr CInt

foreign import ccall unsafe "& H5E_BADATOM_g"       minor_badatom_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_BADGROUP_g"      minor_badgroup_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTREGISTER_g"  minor_cantregister_g  :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTINC_g"       minor_cantinc_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTDEC_g"       minor_cantdec_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_NOIDS_g"         minor_noids_g         :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTFLUSH_g"     minor_cantflush_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSERIALIZE_g" minor_cantserialize_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTLOAD_g"      minor_cantload_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_PROTECT_g"       minor_protect_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_NOTCACHED_g"     minor_notcached_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_SYSTEM_g"        minor_system_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTINS_g"       minor_cantins_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTRENAME_g"    minor_cantrename_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTPROTECT_g"   minor_cantprotect_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTUNPROTECT_g" minor_cantunprotect_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTPIN_g"       minor_cantpin_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTUNPIN_g"     minor_cantunpin_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTMARKDIRTY_g" minor_cantmarkdirty_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTDIRTY_g"     minor_cantdirty_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTEXPUNGE_g"   minor_cantexpunge_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTRESIZE_g"    minor_cantresize_g    :: Ptr CInt

foreign import ccall unsafe "& H5E_TRAVERSE_g"      minor_traverse_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_NLINKS_g"        minor_nlinks_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_NOTREGISTERED_g" minor_notregistered_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTMOVE_g"      minor_cantmove_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSORT_g"      minor_cantsort_g      :: Ptr CInt

foreign import ccall unsafe "& H5E_MPI_g"           minor_mpi_g           :: Ptr CInt
foreign import ccall unsafe "& H5E_MPIERRSTR_g"     minor_mpierrstr_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTRECV_g"      minor_cantrecv_g      :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTCLIP_g"      minor_cantclip_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCOUNT_g"     minor_cantcount_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSELECT_g"    minor_cantselect_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTNEXT_g"      minor_cantnext_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_BADSELECT_g"     minor_badselect_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTCOMPARE_g"   minor_cantcompare_g   :: Ptr CInt

foreign import ccall unsafe "& H5E_UNINITIALIZED_g" minor_uninitialized_g :: Ptr CInt
foreign import ccall unsafe "& H5E_UNSUPPORTED_g"   minor_unsupported_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_BADTYPE_g"       minor_badtype_g       :: Ptr CInt
foreign import ccall unsafe "& H5E_BADRANGE_g"      minor_badrange_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_BADVALUE_g"      minor_badvalue_g      :: Ptr CInt

foreign import ccall unsafe "& H5E_NOTFOUND_g"      minor_notfound_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_EXISTS_g"        minor_exists_g        :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTENCODE_g"    minor_cantencode_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTDECODE_g"    minor_cantdecode_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSPLIT_g"     minor_cantsplit_g     :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTREDISTRIBUTE_g" minor_cantredistribute_g :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTSWAP_g"      minor_cantswap_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTINSERT_g"    minor_cantinsert_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTLIST_g"      minor_cantlist_g      :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTMODIFY_g"    minor_cantmodify_g    :: Ptr CInt
foreign import ccall unsafe "& H5E_CANTREMOVE_g"    minor_cantremove_g    :: Ptr CInt

foreign import ccall unsafe "& H5E_CANTCONVERT_g"   minor_cantconvert_g   :: Ptr CInt
foreign import ccall unsafe "& H5E_BADSIZE_g"       minor_badsize_g       :: Ptr CInt

