{-# LANGUAGE ForeignFunctionInterface #-}
module HDF5.Lite (make_dataset,
                  make_scalar,
                  read_dataset,
                  read_scalar)
where

import Foreign
import Foreign.C
import Data.Array.Storable
import Control.Monad (liftM,when)

import HDF5.Base
import qualified HDF5.Error as H5E
import qualified HDF5.Type as H5T

foreign import ccall unsafe "hdf5.h H5LTmake_dataset"
--               loc_id    name      rank      hsize_t    type_id  buffer
  make_dataset_ :: Lid -> CString -> CInt -> Ptr CULLong -> Tid -> Ptr () -> IO Herr

make_dataset :: (Storable a,H5T.Typed a) => Lid -> String -> StorableArray Hsize a -> IO ()
make_dataset loc name arr = do
  bounds <- getBounds arr
  withCString name $ \cname ->
    -- In HDF5 the last index varies fastest, so we reverse
    -- (this way the Octave HDF5 loader gets the dimensions right, too)
    withArrayLen (reverse $ dimensions bounds) $ \len sizes ->
    withStorableArray arr $ \buffer ->
    H5T.withTypeP buffer $ \tid ->
    H5E.throwIfErr_ ("making "++name) $
      make_dataset_ loc cname (fromIntegral len) sizes tid (castPtr buffer)

make_scalar ::  (Storable a,H5T.Typed a) => Lid -> String -> a -> IO ()
make_scalar loc name e =
  withCString name $ \cname ->
  with e $ \eP ->
  H5T.withType e $ \tid ->
  H5E.throwIfErr_ ("making scalar "++name) $
    make_dataset_ loc cname 0 nullPtr tid (castPtr eP)

foreign import ccall unsafe "hdf5.h H5LTget_dataset_ndims"
--                     loc_id    name      rank
  get_dataset_ndims_ :: Lid -> CString -> Ptr CInt -> IO Herr

get_dataset_ndims :: Lid -> String -> IO Int
get_dataset_ndims loc name =
  alloca $ \rankP ->
  withCString name $ \cname -> do
    H5E.throwIfErr_ ("getting rank of "++name) $ get_dataset_ndims_ loc cname rankP
    fromIntegral `liftM` peek rankP

foreign import ccall unsafe "hdf5.h H5LTget_dataset_info"
--                   loc_id     name      hsize_t       class_id    type_size
  get_dataset_info_ :: Lid -> CString -> Ptr CULLong -> Ptr CInt -> Ptr CSize -> IO Herr

get_dataset_info :: Lid -> String -> IO ((Hsize,Hsize),H5T.Class,CSize)
get_dataset_info loc name = do
  rank <- get_dataset_ndims loc name
  alloca $ \classIdP ->
    alloca $ \typesizeP ->
    allocaArray rank $ \hsizeP -> do
      H5E.throwIfErr_ ("getting info on "++name) $
        withCString name $ \cname -> get_dataset_info_ loc cname hsizeP classIdP typesizeP
      classId <- peek classIdP
      typesize <- peek typesizeP
      hsize <- peekArray rank hsizeP
      return ((Hsize $ replicate rank 1,Hsize $ reverse hsize),H5T.Class classId,typesize)

foreign import ccall unsafe "hdf5.h H5LTread_dataset"
--               loc_id    name    type_id  buffer
  read_dataset_ :: Lid -> CString -> Tid -> Ptr () -> IO Herr

read_dataset :: (Storable a,H5T.Typed a) => Lid -> String -> IO (StorableArray Hsize a)
read_dataset loc name = do
  (bounds,_,_) <- get_dataset_info loc name
  arr <- newArray_ bounds
  withCString name $ \cname ->
    withStorableArray arr $ \buffer ->
    H5T.withTypeP buffer $ \tid -> do
      H5E.throwIfErr_ ("reading "++name) $
        read_dataset_ loc cname tid (castPtr buffer)
      return arr

read_scalar :: (Storable a,H5T.Typed a) => Lid -> String -> IO a
read_scalar loc name = do
  rank <- get_dataset_ndims loc name
  when (rank /= 0) $ ioError $ userError $ "not a scalar: "++name++" has rank "++show rank
  withCString name $ \cname ->
    alloca $ \eP ->
    H5T.withTypeP eP $ \tid -> do
      H5E.throwIfErr_ ("reading scalar "++name) $
        read_dataset_ loc cname tid (castPtr eP)
      peek eP
