{-# LANGUAGE ForeignFunctionInterface #-}
module HDF5.Type (Class(..),
                  Typed(..),
                  withTypeP,
                  create,
                  copy,
                  close,
                  insert,
                  native_int,
                  native_float)
where

import Foreign
import Foreign.C
import Control.Exception (bracket)
import Data.Complex (Complex((:+)),realPart)

import HDF5.Base
import qualified HDF5.Error as H5E

class Typed a where
  withType :: a -> (Tid -> IO b) -> IO b

withTypeP :: Typed a => Ptr a -> (Tid -> IO b) -> IO b
withTypeP = withType . undefinedFromPtr

undefinedFromPtr :: Ptr a -> a
undefinedFromPtr = undefined

instance Typed CInt where
  withType x action = action native_int

instance Typed CFloat where
  withType x action = action native_float

instance (RealFloat a,Typed a,Storable a) => Typed (Complex a) where
  withType x action =
    bracket (create compound size)
            close $ \c -> withType (realPart x) $ \tid -> do
      insert c "real"    0 tid
      insert c "imag" offs tid
      action c
    where size = fromIntegral $ sizeOf x
          offs = size `div` 2

foreign import ccall unsafe "hdf5.h H5Tcreate"
  create_ :: Class -> CSize -> IO Tid

create :: Class -> CSize -> IO Tid
create cls size =
  H5E.throwIfErr "creating datatype" $ create_ cls size

foreign import ccall unsafe "hdf5.h H5Tcopy"
  copy_ :: Tid -> IO Tid

copy :: Tid -> IO Tid
copy =
  H5E.throwIfErr "copying datatype" . copy_

foreign import ccall unsafe "hdf5.h H5Tclose"
  close_ :: Tid -> IO Herr

close :: Tid -> IO ()
close =
  H5E.throwIfErr_ "closing datatype" . close_

foreign import ccall unsafe "hdf5.h H5Tinsert"
--         dtype_id   name     offset field_t
  insert_ :: Tid -> CString -> CSize -> Tid -> IO Herr

insert :: Tid -> String -> CSize -> Tid -> IO ()
insert dtype_id name offset field_t =
  withCString name $ \cname ->
  H5E.throwIfErr_ ("inserting field"++name) $ insert_ dtype_id cname offset field_t

instance (RealFloat a,Storable a) => Storable (Complex a) where
    sizeOf ~(a:+_)    = 2 * sizeOf a
    alignment ~(a:+_) = alignment a
    peek cP           = do [a,b] <- peekArray 2 (castPtr cP)
                           return (a:+b)
    poke cP (a:+b)    = pokeArray (castPtr cP) [a,b]

newtype Class = Class CInt deriving Show

integer    = Class  0           -- integer types
float      = Class  1           -- floating-point types
time       = Class  2           -- date and time types
string     = Class  3           -- character string types
bitfield   = Class  4           -- bit field types
opaque     = Class  5           -- opaque types
compound   = Class  6           -- compound types
reference  = Class  7           -- reference types
enum       = Class  8           -- enumeration types
vlen       = Class  9           -- Variable-Length types
array      = Class 10           -- Array types

newtype Order = Order CInt

order_le   = Order 0            -- little endian
order_be   = Order 1            -- big endian
order_vax  = Order 2            -- VAX mixed endian
order_none = Order 3            -- no particular order (strings, bits,..)

peekTid = Tid . unsafePerformIO . peek

native_char     = native_schar
native_schar    = peekTid native_schar_g
native_uchar    = peekTid native_uchar_g
native_short    = peekTid native_short_g
native_ushort   = peekTid native_ushort_g
native_int      = peekTid native_int_g
native_uint     = peekTid native_uint_g
native_long     = peekTid native_long_g
native_ulong    = peekTid native_ulong_g
native_llong    = peekTid native_llong_g
native_ullong   = peekTid native_ullong_g
native_float    = peekTid native_float_g
native_double   = peekTid native_double_g
native_ldouble  = peekTid native_ldouble_g
native_b8       = peekTid native_b8_g
native_b16      = peekTid native_b16_g
native_b32      = peekTid native_b32_g
native_b64      = peekTid native_b64_g
native_opaque   = peekTid native_opaque_g
native_haddr    = peekTid native_haddr_g
native_hsize    = peekTid native_hsize_g
native_hssize   = peekTid native_hssize_g
native_herr     = peekTid native_herr_g
native_hbool    = peekTid native_hbool_g

ieee_f32be      = peekTid ieee_f32be_g
ieee_f32le      = peekTid ieee_f32le_g
ieee_f64be      = peekTid ieee_f64be_g
ieee_f64le      = peekTid ieee_f64le_g

std_i8be        = peekTid std_i8be_g
std_i8le        = peekTid std_i8le_g
std_i16be       = peekTid std_i16be_g
std_i16le       = peekTid std_i16le_g
std_i32be       = peekTid std_i32be_g
std_i32le       = peekTid std_i32le_g
std_i64be       = peekTid std_i64be_g
std_i64le       = peekTid std_i64le_g
std_u8be        = peekTid std_u8be_g
std_u8le        = peekTid std_u8le_g
std_u16be       = peekTid std_u16be_g
std_u16le       = peekTid std_u16le_g
std_u32be       = peekTid std_u32be_g
std_u32le       = peekTid std_u32le_g
std_u64be       = peekTid std_u64be_g
std_u64le       = peekTid std_u64le_g
std_b8be        = peekTid std_b8be_g
std_b8le        = peekTid std_b8le_g
std_b16be       = peekTid std_b16be_g
std_b16le       = peekTid std_b16le_g
std_b32be       = peekTid std_b32be_g
std_b32le       = peekTid std_b32le_g
std_b64be       = peekTid std_b64be_g
std_b64le       = peekTid std_b64le_g
std_ref_obj     = peekTid std_ref_obj_g
std_ref_dsetreg = peekTid std_ref_dsetreg_g

-- Raw address import for the above
foreign import ccall unsafe "& H5T_NATIVE_SCHAR_g"    native_schar_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_UCHAR_g"    native_uchar_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_SHORT_g"    native_short_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_USHORT_g"   native_ushort_g   :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_INT_g"      native_int_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_UINT_g"     native_uint_g     :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_LONG_g"     native_long_g     :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_ULONG_g"    native_ulong_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_LLONG_g"    native_llong_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_ULLONG_g"   native_ullong_g   :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_FLOAT_g"    native_float_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_DOUBLE_g"   native_double_g   :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_LDOUBLE_g"  native_ldouble_g  :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_B8_g"       native_b8_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_B16_g"      native_b16_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_B32_g"      native_b32_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_B64_g"      native_b64_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_OPAQUE_g"   native_opaque_g   :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_HADDR_g"    native_haddr_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_HSIZE_g"    native_hsize_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_HSSIZE_g"   native_hssize_g   :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_HERR_g"     native_herr_g     :: Ptr CInt
foreign import ccall unsafe "& H5T_NATIVE_HBOOL_g"    native_hbool_g    :: Ptr CInt
foreign import ccall unsafe "& H5T_IEEE_F32BE_g"      ieee_f32be_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_IEEE_F32LE_g"      ieee_f32le_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_IEEE_F64BE_g"      ieee_f64be_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_IEEE_F64LE_g"      ieee_f64le_g      :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I8BE_g"        std_i8be_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I8LE_g"        std_i8le_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I16BE_g"       std_i16be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I16LE_g"       std_i16le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I32BE_g"       std_i32be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I32LE_g"       std_i32le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I64BE_g"       std_i64be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_I64LE_g"       std_i64le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U8BE_g"        std_u8be_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U8LE_g"        std_u8le_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U16BE_g"       std_u16be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U16LE_g"       std_u16le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U32BE_g"       std_u32be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U32LE_g"       std_u32le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U64BE_g"       std_u64be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_U64LE_g"       std_u64le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B8BE_g"        std_b8be_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B8LE_g"        std_b8le_g        :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B16BE_g"       std_b16be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B16LE_g"       std_b16le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B32BE_g"       std_b32be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B32LE_g"       std_b32le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B64BE_g"       std_b64be_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_B64LE_g"       std_b64le_g       :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_REF_OBJ_g"     std_ref_obj_g     :: Ptr CInt
foreign import ccall unsafe "& H5T_STD_REF_DSETREG_g" std_ref_dsetreg_g :: Ptr CInt
